#include <grid/tria.h>
#include <dofs/dof_handler.h>
#include <grid/grid_generator.h>
#include <grid/tria_accessor.h>
#include <grid/tria_iterator.h>
#include <grid/tria_boundary_lib.h>
#include <dofs/dof_accessor.h>
#include <grid/grid_out.h>
#include <fe/fe_q.h>
#include <dofs/dof_tools.h>
#include <fe/fe_values.h>
#include <base/quadrature_lib.h>
#include <base/function.h>
#include <numerics/vectors.h>
#include <numerics/matrices.h>

#include <lac/vector.h>
#include <lac/full_matrix.h>
#include <lac/sparse_matrix.h>
#include <lac/solver_cg.h>
#include <lac/vector_memory.h>
#include <lac/precondition.h>

#include <numerics/data_out.h>
#include <fstream>
#include <iostream>

#include <cstdlib>

#include "../include/Laplace_variations.h"
#include "../include/Output_for_MatLab_2D.h"

#include "../include/Right_Hand_Side.h"
#include "../include/Boundary_Values.h"

using namespace dealii;

/*
	Hierbei handelt es sich um einen Konstruktor mit einer sogenannten Initialisierungsliste.
	D.h. einige der im Header spezifizierten Variablen werden durch den Aufruf ihrer Konstruktoren
	initialisiert.
*/
Laplace_variations::Laplace_variations () :
		fe (1),
		dof_handler (triangulation)
{
	//leer
}



/*
	Tool, welches die Klasse eines gegebenen Zeigers ermittelt und ausgibt.
	Manchmal ganz ntzlich, da es oft vorkommt, dass man nicht wei, was fr ein Typ vorliegt
*/
void Laplace_variations::show_type_of_pointer(void* object){
	std::cout<<std::endl<<"Type is: "<<typeid(object).name()<<std::endl;
}


/*
	Diese Methode erzeugt ein ringfrmiges Gitter und ist weitestgehend aus Tutorial 02 bernommen
*/
void Laplace_variations::make_hyper_shell_grid (Triangulation<2> &tria, int refinement_level){
	// Mittelpunkt des Rings
	const Point<2> center (1,0);
	const double inner_radius = 0.5,
				 outer_radius = 1.0;
	// Vorgefertigte Funktion von deal.II, welche eine Ringstruktur erstellt
	GridGenerator::hyper_shell (tria, center, inner_radius, outer_radius, 10);
    
	// Um bei der Verfeinerung (des bisher groben Gitters) einen runden Rand zu erhalten
	// mssen boundary-objects erstellt werden und der Triangulation bermittelt werden.
	// Das erste Argument (0) heit dabei, dass das boundary-object nur beim Verfeinern von Kanten mit 
	// Boundary-indicator 0 benutzt wird.
	const HyperShellBoundary<2> boundary_description(center);
	tria.set_boundary (0, boundary_description);
  
	// Spezielle Verfeinerung: nach innen wird strker verfeinert (Aus Tut 02 bernommen)                              
	for (int step=0; step < refinement_level; ++step){
			                                       
		Triangulation<2>::active_cell_iterator cell, endc;
		cell = tria.begin_active();
		endc = tria.end();
			      
		for (; cell!=endc; ++cell)
			for (unsigned int vertex=0;
				vertex < GeometryInfo<2>::vertices_per_cell;
				++vertex)
			{
			                                        
			const Point<2> vector_to_center = (cell->vertex(vertex) - center);
			const double distance_from_center = std::sqrt(vector_to_center.square());
			            
			if (std::fabs(distance_from_center - inner_radius) < 1e-10){
				cell->set_refine_flag ();
			    break;
			}
		}
		// Fhrt Verfeinerung letztendlich aus      
		tria.execute_coarsening_and_refinement ();
			      
	}
}

/*    
Erzeugt ein Gitter der Form und speichert es in gegebenen Triangulationobj

	  *
     /|
*---* *
|   |/
*---*


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
12.10.2007 by Vita:
attention: in Version 6 numbering of vertices is different as
in Version 5, when initializing cells
for a cell
*--*
|  |
*--*

it is

3--4
|  |
1--2

so the following two routines are different from Version 5
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

*/
void Laplace_variations::make_two_boxes_grid (Triangulation<2> &tria)
{
	//Quadrat mit folgenden Ecken soll Zelle 0 sein
	Point<2> cell0_p1(0,0);
	Point<2> cell0_p2(1,0);
	Point<2> cell0_p3(1,1);
	Point<2> cell0_p4(0,1);
  
	//Parallelogram mit folgenden Ecken soll Zelle 1 sein
	Point<2> cell1_p1(1,0);
	Point<2> cell1_p2(2,0.5);
	Point<2> cell1_p3(2,1.5);
	Point<2> cell1_p4(1,1);
	
	//Alle Ecken der Zellen werden in einen Vektor gespeichert
	std::vector<Point<2> > vertices_for_both_cells (6);
	
	vertices_for_both_cells[0] = cell0_p1;
	vertices_for_both_cells[1] = cell0_p2;
	vertices_for_both_cells[2] = cell0_p3;
	vertices_for_both_cells[3] = cell0_p4;
	
	vertices_for_both_cells[4] = cell1_p2;
	vertices_for_both_cells[5] = cell1_p3;
	
	//Die folgenden Zeilen werden nicht ausgefhrt, da sonst die Linie
	//zwischen Zelle 1 und 2 von der Methode 'create_triangulation(..)' als Rand betrachtet wird
	//vertices_for_both_cells[6] = cell1_p1;
	//vertices_for_both_cells[7] = cell1_p4;	    
	// Das bedeutet, dass man auf Mehrfachnennungen im Vektor 'vertices_for_both_cells' verzichten sollte.
		
			  

	//Dieser Vektor beinhaltet die Zellen welche gebildet werden sollen
	std::vector<CellData<2> > cells (2);
	
	
	// Zelle 0
	cells[0].vertices[0] = 0;
	cells[0].vertices[1] = 1;
	cells[0].vertices[2] = 3;
	cells[0].vertices[3] = 2;
	cells[0].material_id = 0;
	
	
	//Zelle 1
	cells[1].vertices[0] = 1;
	cells[1].vertices[1] = 4;
	cells[1].vertices[2] = 2;
	cells[1].vertices[3] = 5;
	cells[1].material_id = 0;
	
	//Hier wird das Gitter bzw. die Zellen erzeugt, indem alles zusammengetragen wird.
	//Was es mit dem Object 'SubCellData()' auf sich hat, hab ich nicht ganz verstanden.
	tria.create_triangulation (vertices_for_both_cells, cells, SubCellData());
	
	
}



/*
	Erzeugt Quadratisches Gitter mit einem Loch in der Mitte. Die Ausmasse entsprechen dabei den Forderungen
	zum ersten Teilziel.
*/

void Laplace_variations::make_grid_with_hole (Triangulation<2> &tria, int refinement_level)
{
	//Aussere Ecken des Quadrates mit Laenge 1
	Point<2> p1(0, 0), p2(0.5, 0), p3(1, 0), p4(1, 0.5), p5(1, 1),
  			 p6(0.5, 1), p7(0, 1), p8(0, 0.5);
  
	//8 Ecken des mittleren herausgeschnittenen Kreises mit Radius 0.25
	double a = (4.0-sqrt(2))/8.0;
	Point<2> c_p1(0.5, 0.25), c_p2(1-a, a), c_p3(0.75, 0.5), c_p4(1-a, 1-a),
  			 c_p5(0.5, 0.75), c_p6(a, 1-a), c_p7(0.25, 0.5), c_p8(a, a);
      
  
  
	//Alle Ecken der Zellen werden in einen Vektor gespeichert
	std::vector<Point<2> > vertices (16);
  
    vertices[0] = p1;
    vertices[1] = p2;
    vertices[2] = p3;
    vertices[3] = p4;
    vertices[4] = p5;
    vertices[5] = p6;
    vertices[6] = p7;
    vertices[7] = p8;
    
    vertices[8] = c_p1;
    vertices[9] = c_p2;
    vertices[10] = c_p3;
    vertices[11] = c_p4;
    vertices[12] = c_p5;
    vertices[13] = c_p6;
    vertices[14] = c_p7;
    vertices[15] = c_p8;
			  

	//Dieser Vektor beinhaltet die 8 Zellen welche gebildet werden sollen
	std::vector<CellData<2> > cells (8);
	
	//Zelle 0
	cells[0].vertices[0] = 0;
	cells[0].vertices[1] = 1;
	cells[0].vertices[2] = 15;
	cells[0].vertices[3] = 8;
	
	
	//Zellen 1-6
	for(int i = 1; i<=6; i++){
		cells[i].vertices[0] = i;
		cells[i].vertices[1] = i+1;
		cells[i].vertices[2] = i+7;
		cells[i].vertices[3] = i+8;
	}
  
	//Zelle 7
	cells[7].vertices[0] = 7;
	cells[7].vertices[1] = 0;
	cells[7].vertices[2] = 14;
	cells[7].vertices[3] = 15;
  
	for(int i = 0; i<16; i++){
		cells[i].material_id = 0;
	}
	
	//Hier wird das Gitter bzw. die Zellen erzeugt, indem alles zusammengetragen wird
	tria.create_triangulation (vertices, cells, SubCellData());
	
	//Nun muss noch ein Boundary-Objekt (aehnlich wie bei Tut02) erstellt werden, damit
	//bei zunehmender Verfeinerung das Loch auch wirklich kreisfoermig wird.
	const Point<2> center_of_inner_cycle (0.5, 0.5);
	const double radius = 0.25;
	//const double radius = 0.5*sqrt(2.0);
	static const HyperBallBoundary<2> boundary_descr(center_of_inner_cycle, radius);
	
	//Iterator zum durchlaufen aller Zellen
	Triangulation<2>::cell_iterator cell, end;
	cell =  tria.begin();
	end = tria.end();
	
	//Stelle alle Kanten des 'Kreises' (der ja noch keiner ist) auf boundary indicator 1
	for(; cell!=end ; cell++){
		cell->face(3)->set_boundary_indicator (1);
	}
	
	
	//Verwende bei Verfeinerung HyperballBoundary fr Kanten mit boundary indicator 1
	tria.set_boundary (1, boundary_descr);
	
	//viermal Gitterverfeinerung
	tria.refine_global (refinement_level);
	
	//Rueckgaengigmachen der Veraenderung der boundary indicators, da spaeer die Randwerte auf
	//Kanten mit boundary indicator 0 angewendet werden
	//cells are built so that 2nd face is always the inside of the "circle"
	for(cell = tria.begin(); cell!=end ; cell++){
		if(cell->face(2)->boundary_indicator() == 1){
			cell->face(2)->set_boundary_indicator(0);
		}		
	}

}


/*
	Diese Methode ermittelt bei einem Benutzerdialog, welche Gitterstruktur dem Laplaceproblem
	zu Grunde liegen soll. Dabei hat man die Auswahl zwischen ein paar von deal.II vorgegeben 
	oder von mir erstellten.
*/				
void Laplace_variations::make_grid_and_dofs ()
{
/////////////////////////////////////////////Auswahl der Triangulation /////////////////////////////////
	std::cout<<"On which triangulation do you want to solve the 2 dimensional "<<std::endl;
	std::cout<<"laplace problem?"<<std::endl;
	std::cout<<"--------------------------------------------------------------"<<std::endl;
	std::cout<<"1) deal.II: hyper_cube"<<std::endl;
	std::cout<<"2) deal.II: hyper_L"<<std::endl;
	std::cout<<"3) deal.II: hyper_shell"<<std::endl;
	std::cout<<"4) deal.II: hyper_cube_slit"<<std::endl;
	std::cout<<"5) two boxes"<<std::endl;
	std::cout<<"6) box with hole"<<std::endl;
	
	//Variable speichert Anzahl der gewnschten Verfeinerungsschritte
	int refinChoice;
	std::cout<<std::endl<<"Please choose: ";
	std::cin>>gridChoice;
	std::cout<<std::endl<<"Refinement level: ";
	std::cin>>refinChoice;
	
	//Hier wird der Auswahl entsprechend das Triangulationsobject belegt
	switch(gridChoice){
  		case 1: 
				GridGenerator::hyper_cube (triangulation, -1, 1);
				triangulation.refine_global (refinChoice);
				break;
  		case 2: 
				GridGenerator::hyper_L (triangulation); 
				triangulation.refine_global (refinChoice);
				break;
  		case 3: 
				make_hyper_shell_grid (triangulation, refinChoice); 
	  			break;
  		case 4: 
				GridGenerator::hyper_cube_slit (triangulation, -1, 1);
				triangulation.refine_global (refinChoice);
				break;
		case 5: 
				make_two_boxes_grid(triangulation);
				triangulation.refine_global (refinChoice);
				break;		
		case 6: 
				make_grid_with_hole(triangulation, refinChoice);
				break;
  		default: 
				GridGenerator::hyper_cube (triangulation, -1, 1);
				triangulation.refine_global (refinChoice);
	

	}
///////////////////////////////////////////////////////////////////////////////////////////////


	dof_handler.distribute_dofs (fe);
				  
	//Ausgabe ber Anzahl von Zellen und Freiheitsgraden
	std::cout<<"Number of active cells: "<<triangulation.n_active_cells()<<std::endl;
	std::cout<<"Number of degrees of freedom: "<<dof_handler.n_dofs()<<std::endl;
	
	//Erstellt eine Belegungsmaske fr eine sparse-Matrix
	sparsity_pattern.reinit (dof_handler.n_dofs(),
			dof_handler.n_dofs(),
			dof_handler.max_couplings_between_dofs());		
	DoFTools::make_sparsity_pattern (dof_handler, sparsity_pattern);
	sparsity_pattern.compress(); //'sparsity_pattern' kann nun nicht mehr verndert werden

	//Initialisiert bzw. allokiert nun Speicher fr folgende Objekte
	system_matrix.reinit (sparsity_pattern);
	solution.reinit (dof_handler.n_dofs());
	system_rhs.reinit (dof_handler.n_dofs());
}






/*
	Diese Methode belegt die Variablen 'system_matrix' und 'system_rhs'
*/				 
void Laplace_variations::assemble_system () 
{
	
				
	//Objekt zur Anwendung der Gauschen Quadraturformel
	QGauss<2>  quadrature_formula(2);
	
	//Dieses Objekt steht als Schnittstelle zur Verfgung, welche Daten ber die Basisfunktionen auf
	//der Triangulation liefert.
	//Diese Update-Flags vestehe ich C++-technisch nicht, so wie es im ausgedruckten Tutorial
	//steht. In der Implentation hier wird es anders gemacht!
	FEValues<2> fe_values (fe, quadrature_formula, 
			UpdateFlags(update_values |
						update_gradients |
						update_q_points |
						update_JxW_values));

				  
	//Abkrzende Schreibweisen, um Sourcecode lesbarer zu gestalten
	const unsigned int   dofs_per_cell = fe.dofs_per_cell;
	const unsigned int   n_q_points    = quadrature_formula.n_quadrature_points;
	
	//Temporrer zur Vefgung gestellter Speicher um lokale Daten zwischenzuspeichern
	FullMatrix<double>   cell_matrix (dofs_per_cell, dofs_per_cell);
	Vector<double>       cell_rhs (dofs_per_cell);
	
	//Temporrer zur Vefgung gestellter Speicher um Assoziation zwischen globalen und lokalen
	//dofs zu speichern
	std::vector<unsigned int> local_dof_indices (dofs_per_cell);
	
				  
	DoFHandler<2>::active_cell_iterator cell = dof_handler.begin_active(),
										endc = dof_handler.end();

	

///////////////////////////////////////// Auswahl der "Rechten Seite" //////////////////////
	
	//Zeiger auf ein Objekt vom Typ Function<2>, welcher im weiteren Programverlauf
	//je nach Benutzerwunsch belegt wird. Das Objekt auf welches dieser Zeiger zeigt
	//soll die rechte Seite der DGL beschreiben.
	Function<2>* right_side;

	unsigned int rhsChoice;
	std::cout<<std::endl;
	std::cout<<"Which right hand side do you want?"<<std::endl;
	std::cout<<"-------------------------------------"<<std::endl;
	std::cout<<"1) Constant function f(x) = 1"<<std::endl;
	std::cout<<"2) 'Teilziel #1'-function f(x) = (x-0.5)(y-0.5)"<<std::endl;
	std::cout<<std::endl;
	std::cout<<"Please choose: ";
	std::cin>>rhsChoice;
	std::cout<<std::endl;
	
	if(rhsChoice == 1){
		right_side = new ConstantFunction<2>(1);
	}
	if(rhsChoice == 2){
		right_side = new Right_Hand_Side();
	}
	
/////////////////////////////////////////////////////////////////////////////////////////////
				      
				      
	//Diese Schleife durchluft alle Zellen der Triangulation
	for (; cell!=endc; ++cell) {
			
		//Stellt das Objekt fe_values auf die aktuelle Zelle ein, damit das Objekt entsprechenden 
		//Daten dazu liefer kann
		fe_values.reinit (cell);
		
		//Lscht temporren Speicher
		cell_matrix = 0;
		cell_rhs = 0;
		
		//Belegung der lokalen Matrix
		for (unsigned int i=0; i<dofs_per_cell; ++i)
			for (unsigned int j=0; j<dofs_per_cell; ++j)
		  		for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
					cell_matrix(i,j) += (fe_values.shape_grad (i, q_point) *
					fe_values.shape_grad (j, q_point) *
					fe_values.JxW (q_point));
		

		
		//Belegung der lokalen rechten Seite
		for (unsigned int i=0; i<dofs_per_cell; ++i)
			for (unsigned int q_point=0; q_point<n_q_points; ++q_point){
				cell_rhs(i) += (fe_values.shape_value (i, q_point) *
					(*right_side).value(fe_values.quadrature_point(q_point)) *
					fe_values.JxW (q_point));
			}
		
		//Ermittelt die den lokalen Ecken zugeordneter globalen Indizies
		// d.h. local_dof_indices[2] enthlt den globalen Index der lokal mit 2 indizierten Ecke
		cell->get_dof_indices (local_dof_indices);
		
		
		//Einbinden der lokal ermittelten Daten in globale Matrix			      
		for (unsigned int i=0; i<dofs_per_cell; ++i)
			for (unsigned int j=0; j<dofs_per_cell; ++j)
		  		system_matrix.add (local_dof_indices[i],
				     local_dof_indices[j],
				     cell_matrix(i,j));
		
		//Einbinden der lokal ermittelten Daten in globale rechte Seite
		for (unsigned int i=0; i<dofs_per_cell; ++i)
			system_rhs(local_dof_indices[i]) += cell_rhs(i);
	
	}
	
	//Der Speicher muss wieder freigegeben werden, da dies im Falle einer Zeigererzeugung
	//nicht automatisch geschieht
	delete right_side;


//////////////////////////////Auswahl der Dirichlet-Randbed. ////////////////////////////////////
	//Gleichungssystem muss noch entsprechend der Dirichlet-Randbed. angepasst werden
	//Hierbei wird auch wieder bzgl. der anfnglichen Auswahl unterschieden
	std::map<unsigned int,double> boundary_values;

	switch(gridChoice){
  		case 1: case 2: case 3: case 4: case 5:
			//Randbedingung global Null
			VectorTools::interpolate_boundary_values (dof_handler, 0, ZeroFunction<2>(),
					    boundary_values);
			break;
			
		case 6: //Box mit Loch
			unsigned int boundaryChoice;
			std::cout<<std::endl;
			std::cout<<"Which boundary condition do you want?"<<std::endl;
			std::cout<<"-------------------------------------"<<std::endl;
			std::cout<<"1) Zero condition"<<std::endl;
			std::cout<<"2) 'Teilziel #1' condition"<<std::endl;
			std::cout<<std::endl;
			std::cout<<"Please choose: ";
			std::cin>>boundaryChoice;
			std::cout<<std::endl;
			
			//Randbedingung global Null
			if(boundaryChoice == 1){
	  			VectorTools::interpolate_boundary_values (dof_handler, 0, ZeroFunction<2>(),  boundary_values);
			}
			//Randbedingung gem formulierten Teilziel #1
			if(boundaryChoice == 2){
				const Boundary_Values bv;
	  			VectorTools::interpolate_boundary_values (dof_handler, 0, bv,  boundary_values);
			}
			break;
  		default: 
				VectorTools::interpolate_boundary_values (dof_handler, 0, ZeroFunction<2>(),
					    boundary_values);
	}
///////////////////////////////////////////////////////////////////////////////////////////////////////
				 
	MatrixTools::apply_boundary_values (boundary_values,
				      system_matrix,
				      solution,
				      system_rhs);
}





/*
	Diese Methode lst das mit der Methode 'assemble_system' erstellte Gleichungssystem und
	speichert die Lsung im globalen Vektor 'solution'.
*/
void Laplace_variations::solve () 
{
	//Abbruchsbedingung: mehr als 1000 Iterationen oder einen Fehler (bzgl. einer Norm) kleiner
	//als 10^(-12)
	SolverControl solver_control (1000, 1e-12);
	
	//Von diesem Vektor steht nichts im Tutorial! Die Version im Tutorial kompiliert nicht.
	// -> no matching function for call to 'SolverCG<Vector<double> >::SolverCG(SolverControl&)' 
	PrimitiveVectorMemory<> vector_memory;
				   
	SolverCG<> cg (solver_control, vector_memory);		  
	cg.solve (system_matrix, solution, system_rhs, PreconditionIdentity());
				   
}


/*
	Diese Methode erstellt Ausgaben fr Gnuplot und als *.eps
*/
void Laplace_variations::output_results () const
{
				  
				  
	DataOut<2> data_out;
				   
	data_out.attach_dof_handler (dof_handler);
	data_out.add_data_vector (solution, "solution");
				   
	data_out.build_patches ();
	
	//Ausgabe als Gnuplot-Datei
	std::ofstream outputGnuPlot ("solution_for_GnuPlot.gpl");
	data_out.write_gnuplot (outputGnuPlot);
	
	
	//Ausgabe fr MatLab
	std::ofstream outputMatLab ("solution_for_MatLab.txt");
	Output_for_MatLab_2D::ready_for_MatLab(outputMatLab, dof_handler, solution);
	

  
	//Ausgabe als *.eps (allerdings nicht klar aus welchem Winkel die Abbildung genommen wird)
	std::ofstream outputEps ("solution.eps");
	data_out.write_eps (outputEps);
	
  std::ofstream out ("grid.eps");
  GridOut grid_out;
  grid_out.write_eps (triangulation, out);	

}


/*
	Diese Methode stellt den konkreten Programablauf dar.
	1. Gittererzeugung
	2. Gleichungssystem erstellen
	3. Lsen
	4. Ausgabe
*/				
void Laplace_variations::run () 
{
	make_grid_and_dofs ();
	assemble_system ();
	solve ();
	output_results ();
}


/*
	Methode hat nur den einen Zweck, die Ausfhung des Programs erst nach drcken der Entertaste
	zu beenden
*/
void wait () 
{ 
	std::cout<<std::endl<<"Press 'Enter' to end program...";
	std::cin.get();
	std::cin.get();
	
} 



int main () 
{
	//Hier "fehlen" eigentlich die Klammer nach 'laplace_problem'! Damit das Objekt
	//mit Hilfe des Standardkonstruktors erzeugt wird. Allerdings fordert der C++-Syntax,
	//dass man die Klammern weglassen MUSS!
	Laplace_variations laplace_problem;
	laplace_problem.run ();
	
	wait();
	 
	return 0;
}
